'''
.. data:: ANY
    TEst test test
'''

from enum import Enum


class BuiltinEnum(Enum):

    ANY = 'Any'
    BOOL = 'Bool'
    FLOAT = 'Float'
    INT = 'Int'
    PATH = 'Path'
    STRING = 'String'
    NOTHING = 'Nothing'
